from pathlib import Path
from typing import Generator, Iterable, Tuple

import click

from jinja2_pdoc import Jinja2Pdoc, jinja2


def eof_newline(content: str, eof: str = "\n") -> str:
    """
    make sure the file content ends with a newline if specified.
    """
    if content.endswith(eof) or not eof:
        return content

    return content + eof


def load_files(
    files: Iterable[Path], out_dir: Path, force: bool, root: Path = None
) -> Generator[Tuple[str, Path], None, None]:
    """
    iterates over files and yield `(content,  out_file)` if its not existing.

    if `force` is True, all files are proessed.
    """
    for file in files:
        if not root:
            out = out_dir.joinpath(file.name)
        else:
            out = out_dir.joinpath(file.relative_to(root))

        if not out.is_file() or force:
            yield (file.read_text(), out.with_suffix(""))
            click.echo(f"rendering.. {out}")
        else:
            click.echo(f"skip....... {out}")
    else:
        click.echo("\n......done")


@click.command()
@click.argument("input", type=click.Path(exists=True))
@click.argument("output", type=click.Path(file_okay=False), default=Path.cwd())
@click.option(
    "-p",
    "--pattern",
    default="*.jinja2",
    help="template search pattern for directories",
)
@click.option("-f", "--force", is_flag=True, help="overwrite existing files")
@click.option(
    "-n",
    "--newline",
    default="\n",
    help="newline character",
)
def main(
    input: str,
    output: str = ".",
    pattern: str = "*.jinja2",
    force: bool = False,
    newline: str = "\n",
) -> None:
    """
    Render jinja2 templates from a input directory or file and
    write to a output directory.

    if the `input` is a directory, all files with a matching `pattern` are renderd.

    if no `output` is given, the current working directory is used.
    """

    env = jinja2.Environment(extensions=[Jinja2Pdoc])

    input = Path(input).resolve()
    output = Path(output).resolve()

    if input.is_file():
        files = [
            input,
        ]
        root = None
    else:
        files = input.rglob(pattern)
        root = input

    for content, file in load_files(files, output, force, root):
        code = env.from_string(content).render()
        file.parent.mkdir(parents=True, exist_ok=True)
        file.write_text(eof_newline(code, newline))


if __name__ == "__main__":
    main()



import re
import textwrap
from functools import cached_property
from pathlib import Path
from typing import Any, Union

import pdoc
import autopep8


class Function(pdoc.doc.Function):
    """
    A wrapper __class__ to cast an incance of `pdoc.doc.Function`
    to enhance it with new methods
    """

    _regex_doc = re.compile(
        r"^\s*?(?P<doc>\"{3}|\'{3}).*?(?P=doc)\s*$", re.MULTILINE | re.DOTALL
    )
    """regex to match a docstring"""

    _regex_def = re.compile(r"def.*?\(.*?\)(?:.*?)?:\s*$", re.MULTILINE | re.DOTALL)
    """regex to match a function definition"""

    def __new__(cls, obj: pdoc.doc.Function) -> Union["Function", None]:
        """
        create a new instance of `Function` with the wrapped object
        """
        if obj is None:
            return obj

        return super().__new__(cls)

    def __init__(self, obj: pdoc.doc.Function) -> None:
        """
        store the object to wrap
        """
        self.__obj = obj

    def __getattribute__(self, name: str) -> Any:
        """
        get all known attributes and cast `str` to `PdocStr`
        """
        attr = super().__getattribute__(name)

        if isinstance(attr, str):
            attr = PdocStr(attr)

        return attr

    def __getattr__(self, name):
        """
        get all unknown attributes from the wrapped object
        """
        attr = getattr(self.__obj, name)

        return attr

    @cached_property
    def code(self) -> "PdocStr":
        """
        returns the source without docstring and function definition
        """
        code = self._regex_def.sub("", self.source, 1)
        code = self._regex_doc.sub("", code, 1)

        return PdocStr(code.strip("\n"))


class Module(pdoc.doc.Module):
    """
    Subclass of `pdoc.doc.Module` to override the `get` method to return a instance of
    `Function` instead of `pdoc.doc.Function`
    """

    def get(self, name: str) -> Union[Function, None]:
        return Function(super().get(name))

    @classmethod
    def from_name(cls, name: str) -> "Module":
        """
        create a `Module` instance from a module name or a file path
        """
        try:
            return super().from_name(name)
        except RuntimeError:
            path = Path(name).with_suffix("")
            name = ".".join(path.parts)

            return super().from_name(name)


class PdocStr(str):
    """
    inhertits from `str` with a `dedent` method
    """

    def dedent(self) -> str:
        """
        remove common whitespace from the left of every line in the string,
        see `textwrap.dedent` for more information.
        """
        s = textwrap.dedent(self)
        return self.__class__(s)

    def indent(self) -> str:
        """
        remove leading spaces and change indent size to 2 spaces, instead of 4.
        """
        s = autopep8.fix_code(self.dedent(), options={"indent_size": 2})
        return self.__class__(s)



import textwrap
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict

import jinja2

from jinja2_pdoc.wrapper import PdocStr, Module


class Jinja2Pdoc(jinja2.ext.Extension):
    tags = {"pdoc"}

    def parse(self, parser: jinja2.parser.Parser) -> jinja2.nodes.Node:
        """
        replace a `{{ pdoc module:name:__attr__ }}` with the source code from a
        the python module. `__attr__` is optional and defaults to `source`, see
        `pdoc.doc.Functions` which attributes are available.
        """
        lineno = next(parser.stream).lineno

        tokens = []
        while parser.stream.current.type != "block_end":
            tokens.append(parser.stream.current.value)
            parser.stream.skip(1)

        arg = "".join(tokens)
        try:
            text = self._pdoc_jinja2(arg)
        except ValueError:
            raise jinja2.TemplateSyntaxError(f"syntax error in '{arg}'", lineno)
        except AttributeError:
            raise jinja2.TemplateAssertionError(f"could not resolve '{arg}'", lineno)

        content = jinja2.nodes.Const(text)

        return jinja2.nodes.Output([content]).set_lineno(lineno)

    @staticmethod
    def _pdoc_syntax(line: str) -> Dict[str, str]:
        """
        Parse a line of the form `module:name:__attr__` and return a dict with
        corresponding keys and values.

        - `module` is the module name or file path
        - `name` is the name of the function or class
        - `attr` is the optional attribute which will be called from the pdoc object.
        Default: `source`

        Example:
        >>> PdocJinja2._pdoc_syntax("pathlib::Path.open")
        {'module': 'pathlib', 'name': 'Path.open', 'attr': 'source'}
        """
        pdoc = {}

        try:
            pdoc["module"], line = line.split(":", 1)
        except ValueError:
            pdoc["module"] = line
            pdoc["name"] = ""
            pdoc["attr"] = "source"
            return pdoc

        try:
            pdoc["name"], line = line.split(":", 1)
        except ValueError:
            pdoc["name"] = line
            pdoc["attr"] = "source"
            return pdoc

        try:
            attr, frmt = line.split(".", 1)
        except ValueError:
            pdoc["attr"] = line.strip("_") or "source"
        else:
            pdoc["attr"] = attr.strip("_") or "source"
            frmt = frmt.strip("_")
            if frmt:
                pdoc["frmt"] = frmt

        return pdoc

    @staticmethod
    @lru_cache
    def _pdoc_load(module: str) -> Module:
        """
        Load a module and return a subclass of `pdoc.doc.Module` instance.
        """
        return Module.from_name(module)

    @classmethod
    def _pdoc_jinja2(cls, line: str) -> PdocStr:
        """
        Return the code segment of a function or class from a module.

        Example:
        >>> PdocJinja2._pdoc_jinja2("pathlib:Path.open:__docstring__")
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        cfg = cls._pdoc_syntax(line)

        doc = cls._pdoc_load(cfg["module"])

        if cfg["name"]:
            s = getattr(doc.get(cfg["name"]), cfg["attr"])
        else:
            s = getattr(doc, cfg["attr"])

        if "frmt" in cfg.keys():
            s = getattr(PdocStr(s), cfg["frmt"])

            if callable(s):
                s = s()

        return PdocStr(s)
